/*
Interfacing the Arduino Duemilanove to the ElecFreaks alphanumeric display
brick.

Copyright (C) 2016 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <Wire.h>

const int BUFSIZE = 17;
char inbuffer[BUFSIZE];
uint8_t nread = 0;

uint8_t data[8] = {0, 0, 0, 0, 0, 0, 0, 0};

void enable_clock()
{
    Wire.beginTransmission(0x70);
    Wire.write(0x21);
    Wire.endTransmission();
}

void set_row_output()
{
    Wire.beginTransmission(0x70);
    Wire.write(0xa0);
    Wire.endTransmission();
}

void set_dimming()
{
    Wire.beginTransmission(0x70);
    Wire.write(0xe3); // pulse_width = 4/16
    Wire.endTransmission();
}

void enable_display()
{
    Wire.beginTransmission(0x70);
    Wire.write(0x81); // no blinking
    Wire.endTransmission();
}

void disable_display()
{
    Wire.beginTransmission(0x70);
    Wire.write(0x80);
    Wire.endTransmission();
}

void init_alphanumeric()
{
    enable_clock();
    set_row_output();
    set_dimming();
}

void write_digits(uint8_t data[], uint8_t len)
{
    uint8_t i;

    Wire.beginTransmission(0x70);
    Wire.write(0x00); // address = 0

    for (i = 0; i < len; i++)
    {
        Wire.write(data[i]);
    }

    Wire.endTransmission();
}

/* User interface functions. */

uint8_t fromHex(char c)
{
    if ((c >= 48) && (c <= 57))
        return c - 48;
    if ((c >= 65) && (c <= 70))
        return c - 65 + 10;
    if ((c >= 97) && (c <= 102))
        return c - 97 + 10;
    return 0;
}

void to_digits(char buffer[], uint8_t data[], uint8_t len)
{
    uint8_t i, j, p, high, low;

    for (i = 0, j = 0; j < len; j += 2)
    {
        for (p = 2; p > 0; p--, i += 2)
        {
            high = fromHex(buffer[i]);
            if (high == 17)
                return;
            low = fromHex(buffer[i+1]);
            if (low == 17)
                return;

            /* Switch to little-endian. */

            data[j+p-1] = (high << 4) + low;
        }
    }
}

void setup()
{
    Wire.begin();
    Serial.begin(115200);

    init_alphanumeric();
    write_digits(data, 8);
    enable_display();

    // Interface loop.

    Serial.println("?");
}

void loop()
{
    /* Read bytes, obtaining the number read excluding any newline terminator. */

    if (nread += Serial.readBytesUntil('\n', inbuffer + nread, BUFSIZE - nread))
    {
        /* Handle each command, waiting for the newline. */

        if (nread >= 16)
        {
            to_digits(inbuffer, data, 8);
            write_digits(data, 8);
            nread = 0;
            Serial.println("OK");
        }
        else
        {
            Serial.print(nread);
            Serial.println("...");
        }

        Serial.flush();
    }
}

extern "C" void __cxa_pure_virtual(void) {
    while(1);
}

// tabstop=4 expandtab shiftwidth=4
